@@ -714,7 +714,16 @@ check_running_in_user_namespace (libcrun_error_t *err)
 
   ret = read_all_file ("/proc/self/uid_map", &buffer, &len, err);
   if (UNLIKELY (ret < 0))
-    return ret;
+    {
+      /* If the file does not exist, then the kernel does not support user namespaces and we for sure aren't in one.  */
+      if (crun_error_get_errno (err) == ENOENT)
+        {
+          crun_error_release (err);
+          run_in_userns = 0;
+          return run_in_userns;
+        }
+      return ret;
+    }
 
   ret = strstr (buffer, "4294967295") ? 0 : 1;
   run_in_userns = ret;
@@ -2528,3 +2537,37 @@ consume_trailing_slashes (char *path)
 
   *last = '\0';
 }
+
+char **
+read_dir_entries (const char *path, libcrun_error_t *err)
+{
+  cleanup_dir DIR *dir = NULL;
+  size_t n_entries = 0;
+  size_t entries_size = 16;
+  char **entries = NULL;
+  struct dirent *de;
+
+  dir = opendir (path);
+  if (UNLIKELY (dir == NULL))
+    {
+      crun_make_error (err, errno, "opendir `%s`", path);
+      return NULL;
+    }
+
+  entries = xmalloc (entries_size * sizeof (char *));
+  while ((de = readdir (dir)))
+    {
+      if (strcmp (de->d_name, ".") == 0 || strcmp (de->d_name, "..") == 0)
+        continue;
+      if (n_entries == entries_size)
+        {
+          entries_size *= 2;
+          entries = xrealloc (entries, entries_size * sizeof (char *));
+        }
+      entries[n_entries++] = xstrdup (de->d_name);
+    }
+  entries = xrealloc (entries, (n_entries + 1) * sizeof (char *));
+  entries[n_entries] = NULL;
+
+  return entries;
+}
